/* This file is part of Mailfromd.
   Copyright (C) 2006, 2007, 2008, 2009 Sergey Poznyakoff

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>. */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif
#include <stdlib.h>
#include <errno.h>
#include "libmf.h"

/* |hash_size| defines a sequence of symbol table sizes. These are prime
   numbers, the distance between each pair of them grows exponentially,
   starting from 64. Hardly someone will need more than 16411 symbols, and
   even if someone will, it is easy enough to add more numbers to the
   sequence. */

static unsigned int hash_size[] = {
    37,   101,  229,  487, 1009, 2039, 4091, 8191, 16411
};

/* |max_rehash| keeps the number of entries in |hash_size| table. */
static unsigned int max_rehash = sizeof (hash_size) / sizeof (hash_size[0]);

void *(*syment_alloc_hook)(size_t) = NULL;

static void *
syment_alloc(size_t size)
{
	if (!syment_alloc_hook)
		return malloc(size);
	return syment_alloc_hook(size);
}
				   
struct symtab {
	int flags;
	unsigned int hash_num;  /* Index to hash_size table */
	size_t elsize;          /* Size of an element */
	struct syment **tab;
	void (*free_syment) (void *);
};

struct _symtab_align {
	char c;
	struct syment x;
};


static unsigned
hash(const char *name, unsigned long hash_num)
{
	unsigned i;
	
	for (i = 0; *name; name++) {
		i <<= 1;
		i ^= *(unsigned char*) name;
	}
	return i % hash_size[hash_num];
}

static void
symtab_free_syment(struct symtab *st, struct syment *elem)
{
	if (st->free_syment)
		st->free_syment(elem->data);
	if (!(st->flags & SYMTAB_COPY_KEY))
		free(elem->name);
	elem->name = NULL;
}

static unsigned
symtab_insert_pos(struct symtab *st, const char *name)
{
	unsigned i;
	unsigned pos = hash(name, st->hash_num);
	
	for (i = pos; st->tab[i];) {
		if (++i >= hash_size[st->hash_num])
			i = 0;
		if (i == pos)
			/* FIXME: Error message? */
			abort();
	}
	return i;
}

#define name_cmp(st,a,b) (((st)->flags & SYMTAB_ICASE) ? \
			  strcasecmp(a,b) : strcmp(a,b))

int
symtab_replace(struct symtab *st, struct syment *ent, struct syment **old_ent)
{
	struct syment *entry;
	const char *name = ent->name;
	unsigned i, pos = hash(name, st->hash_num);
	for (i = pos; entry = st->tab[i];) {
		if (name_cmp(st, entry->name, name) == 0)
			break;
		if (++i >= hash_size[st->hash_num])
			i = 0;
		if (i == pos)
			return ENOENT;
	}
	if (old_ent)
		*old_ent = entry;
	st->tab[i] = ent;
	return 0;
}

static int
symtab_rehash(struct symtab *st)
{
	struct syment **old_tab = st->tab;
	struct syment **new_tab;
	unsigned int i;
	unsigned int hash_num = st->hash_num + 1;
	
	if (hash_num >= max_rehash)
		return E2BIG;

	new_tab = calloc(hash_size[hash_num], sizeof(*new_tab));
	if (!symtab_rehash)
		return ENOMEM;
	st->tab = new_tab;
	if (old_tab) {
		st->hash_num = hash_num;
		for (i = 0; i < hash_size[hash_num-1]; i++) {
			struct syment *elt = old_tab[i];
			if (elt->name) {
				unsigned n = symtab_insert_pos(st, elt->name);
				new_tab[n] = elt;
			}
		}
		free(old_tab);
	}
	return 0;
}

const char *
symtab_strerror(int rc)
{
	switch (rc) {
	case ENOENT:
		return _("element not found in table");
	case E2BIG:
		return _("symbol table is full");
	case ENOMEM:
		return strerror(ENOMEM);
	}
	return strerror(rc);
}

int
symtab_remove(struct symtab *st, const char *name)
{
	unsigned int pos, i, j, r;
	struct syment *entry;
	
	pos = hash(name, st->hash_num);
	for (i = pos; entry = st->tab[i];) {
		if (name_cmp(st, entry->name, name) == 0)
			break;
		if (++i >= hash_size[st->hash_num])
			i = 0;
		if (i == pos)
			return ENOENT;
	}
	
	symtab_free_syment(st, entry);

	for (;;) {
		st->tab[i] = NULL;
		j = i;

		do {
			if (++i >= hash_size[st->hash_num])
				i = 0;
			if (!st->tab[i])
				return 0;
			r = hash(st->tab[i]->name, st->hash_num);
		}
		while ((j < r && r <= i)
		       || (i < j && j < r) || (r <= i && i < j));
		st->tab[j] = st->tab[i];
	}
	return 0;
}

int
symtab_lookup_or_install(struct syment **elp,
			 struct symtab *st, const char *name, int *install)
{
	int rc;
	unsigned i, pos;
	struct syment *elem;
  
	if (!st->tab) {
		if (install) {
			rc = symtab_rehash(st);
			if (rc)
				return rc;
		} else
			return ENOENT;
	}

	pos = hash(name, st->hash_num);

	for (i = pos; elem = st->tab[i];) {
		if (name_cmp(st, elem->name, name) == 0) {
			if (install)
				*install = 0;
			*elp = elem;
			return 0;
		}
      
		if (++i >= hash_size[st->hash_num])
			i = 0;
		if (i == pos)
			break;
	}

	if (!install)
		return ENOENT;
  
	if (!elem) {
		*install = 1;
		st->tab[i] = syment_alloc(st->elsize);
		if (!st->tab[i])
			return ENOMEM;
		elem = st->tab[i];
		if (st->flags & SYMTAB_COPY_KEY)
			elem->name = (char *) name;
		else {
			elem->name = strdup(name);
			if (!elem->name)
				return ENOMEM;
		}
		*elp = elem;
		return 0; 
	}

	if ((rc = symtab_rehash(st)) != 0)
		return rc;

	return symtab_lookup_or_install(elp, st, name, install);
}

void
symtab_clear(struct symtab *st)
{
	unsigned i, hs;
  
	if (!st || !st->tab)
		return;

	hs = hash_size[st->hash_num];
	for (i = 0; i < hs; i++) {
		struct syment *elem = st->tab[i];
		if (elem) {
			symtab_free_syment(st, elem);
			st->tab[i] = NULL;
		}
	}
}

struct symtab *
symtab_create(size_t elsize, int flags, void (*free_fun)(void *))
{
	struct symtab *st = calloc(1, sizeof (*st));
	if (st) {
		st->flags = flags;
		st->elsize = elsize;
		st->free_syment = free_fun;
	}
	return st;
}

void
symtab_free(struct symtab **pst)
{
	if (pst && *pst) {
		struct symtab *st = *pst;
		symtab_clear(st);
		free(st->tab);
		free(st);
		*pst = NULL;
	}
}

int
symtab_enumerate(struct symtab *st, symtab_enumerator_t fun, void *data)
{
	unsigned i;
	
	for (i = 0; i < hash_size[st->hash_num]; i++) {
		struct syment *ep = st->tab[i];
		if (ep) {
			int rc = fun(ep, data);
			if (rc)
				return rc;
		}
	}
	return 0;
}










   
